//  Data Structures with Java by John R. Hubbard
//  Copyright McGraw-Hill, 2001
//  An AVLTree class

public class AVLTree extends BinaryTree
{ protected AVLTree left, right;
  protected int balance;
  protected java.util.Comparator comp;
  
  public AVLTree(java.util.Comparator comp)
  { this.comp = comp;
  }

  public AVLTree(Object root, java.util.Comparator comp)
  { this.root = root;
    this.comp = comp;
  }

  public boolean add(Object object)
  { AVLTree temp = attach(object);
    if (temp != this)
    { left = temp.left;
      right = temp.right;
      balance = temp.balance;
    }
    return true;
  }

  public AVLTree attach(Object object)
  { if (root == null)  // tree is empty
    { root = object;
      return this;
    }
    if (comp.compare(object,root)<0)  // insert into left subtree
    { if (left == null)
      { left = new AVLTree(object,comp);
        ++size;
        --balance;
      }
      else
      { int lb = left.balance;
        left = left.attach(object);
        if (left.balance != lb && left.balance != 0) --balance;
      }
      if (balance < -1)
      { if (left.balance > 0) left = left.rotateLeft();
        return rotateRight();
      }
    }
    else  // insert into right subtree
    { if (right == null)
      { right = new AVLTree(object,comp);
        ++size;
        ++balance;
      }
      else
      { int rb = right.balance;
        right = right.attach(object);
        if (right.balance != rb && right.balance != 0) ++balance;
      }
      if (balance > 1) 
      { if (right.balance < 0) right = right.rotateRight();
        return rotateLeft();
      }
    }
    return this;
  }

  private AVLTree rotateRight()
  { // PRECONDITION: this.left != null;
    AVLTree x = this;
    AVLTree y = left;
    AVLTree z = y.left;
    x.left = z;
    y.left = x;
    int xb = x.balance;
    int yb = y.balance;
    if (yb > 1)
    { ++x.balance;
      y.balance = ( xb<0 ? yb+1 : xb+yb+2 );
    }
    else if (yb > xb)
    { x.balance += yb-1;
      ++y.balance;
    }
    else y.balance = xb+2;
    return y;
  }
  
  private AVLTree rotateLeft()
  { // PRECONDITION: this.right != null;
    AVLTree x = this;
    AVLTree y = right;
    AVLTree z = y.left;
    x.right = z;
    y.left = x;
    int xb = x.balance;
    int yb = y.balance;
    if (yb < 1)
    { --x.balance;
      y.balance = ( xb>0 ? yb-1 : xb+yb-2 );
    }
    else if (yb < xb)
    { x.balance -= yb+1;
      --y.balance;
    }
    else y.balance = xb-2;
    return y;
  }
}
